/*
 *    Computes theta for the proximal TR orientation, and stores it in
 *       the scene file.
 *
 *    TRx and TRy are used as x,y offsets for the image here.  In fact,
 *       they're x,y offsets to the center of the image, rather than its
 *       upper left corner.
 */

#include <stdio.h>
#include <dirent.h>
#include <sys/types.h>
#include <math.h>
#include "X11/Intrinsic.h"
#include "X11/StringDefs.h" 
#include "X11/Shell.h" 
#include "X11/Xaw/Command.h"
#include "X11/Xaw/Label.h"
#include "X11/Xaw/Box.h"
#include "X11/Xaw/Cardinals.h"
#include "X11/Xaw/Dialog.h"
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>

#define PI    3.1415927
#define FALSE 0
#define TRUE 1

#define DIM 50             /* Object image is DIMxDIM. */
#define SDIM 150           /* Scene is SDIMxSDIM */

#define N 100              /* Max number of units allowed in net */
#define NLMS 15            /* Max number of landmarks used in training */
#define MAXTHRU 10         /* Max number units going through points */
#define MAXTIED 20         /* Max number of classes of tied weights */
#define MAXFIXED 20        /* Max number of fixed weights allowed */
#define MAXSCENES 30       /* Max number of scenes to train on. */
#define MAXFRAMES 100       /* Max number of frames in a movie. */

/*  X-related global variables */

Display              *disptr;
Window               win; 
GC                   gc;
XEvent               event;
XImage               *AnImage;

typedef struct
{
  float Sine;
  float Cosine;
} Angle;

/* Object stuff */

int NumFrames;
int TRx[MAXFRAMES],TRy[MAXFRAMES], LMx,LMy;
char CurObjfname[80];
XImage CurObjimage;
XImage TRimage;
XImage LMimage;
XImage TRmarkimage;
XImage LMmarkimage;
XImage FilledTRimage;
XImage FilledLMimage;

Angle ProxOrientation[MAXFRAMES];     /* LM to TR ray */
int   Spread[SDIM][SDIM];   /* Spread map */
int   LMClosestX, LMClosestY; /* LM point closest to TR */
int   TRClosestX, TRClosestY; /* TR point closest to LM */

int   NumLMBoundaryPoints;
int   NumTRBoundaryPoints;
int   LMBoundaryPoints[DIM*DIM][2];
int   TRBoundaryPoints[MAXFRAMES][DIM*DIM][2];

int X_off, Y_off;

char LMfilename[80], TRfilename[80];

/*  Xt stuff, see apology below ("kind of gross").  */

Widget toplevel;
static XrmOptionDescRec options[] = 
{
    {"-hspace",	"*Box.hSpace",	XrmoptionSepArg,	NULL},
    {"-vspace",	"*Box.vSpace",	XrmoptionSepArg,	NULL},
};

main(argc,argv)
     int argc;
     char *argv[];
{
  int ing;                  /* Needed for command line parsing */
  extern char *optarg;      /* Needed for command line parsing */

  int i, j, k, epox, response;
  char fname[80];
  char SCfname[80];
  char pos_res_fname[80];         /* File for results from +ive instances */
  char neg_res_fname[80];         /* File for results from -ive instances */
  char cmdname[80];
  char inc;
  FILE  *pos_res_file, *neg_res_file, *fopen();

  int status;
  unsigned int *width_ret, *height_ret;
  int *x_hot_ret, *y_hot_ret;

  int x,y, xc,yc;
  float sum, sigma, Gaussian();

  int fm;    /* Frame number */

  int max_tr_ix, max_lm_ix;
  float max_tr_sum, max_lm_sum, tmp_sum;

  XtAppContext app_con;
  Angle *directionptr;

  SCfname[0] = '*';

  directionptr = (Angle *) malloc (sizeof(Angle));

  /* Read in options */

  while ((ing = getopt(argc,argv,"s:")) != EOF)
    switch (ing)
      {
      case 's': strcpy(SCfname,optarg);       break;
      }

  if (SCfname[0] == '*')
    {
      fprintf(stderr, ">> Error:  No scene file specified.\n");
      fprintf(stderr, "   get_theta -s <scenefile>\n");
      exit(1);
    }


  /*
   *  Kind of gross.  Had to have this guy in here just so I could use
   *    the XmuReadBitmapFile routine (see below).   It's not a widget
   *    based program at all.
   */

  toplevel = XtAppInitialize(&app_con, "Xbuttonbox",
			     options, XtNumber(options),
			     &argc, argv, NULL, NULL, ZERO);


  /*  Display initialization (but don't map). */

  disptr=XOpenDisplay(NULL);
  win=XCreateSimpleWindow(disptr,RootWindow(disptr,0), 50,100,
			  175,175, 5,0,  WhitePixel(disptr,0));
  AnImage = XCreateImage(disptr,&win,
			 1, XYBitmap, 0, NULL, DIM,DIM,
			 8, ((DIM + 7) / 8));


  status = ReadScreenFile(SCfname);
  if (status)
    {

      for (fm=0; fm<NumFrames; fm++)
	{

	  /* Find the TR point with the max Gaussian (maxed over LM points).
	   *   I.e. find the TR point for which the max Gaussian to any LM
	   *   point is max over all other TR points.
	   */

	  max_tr_ix = 0;
	  max_tr_sum = 0.0;
	  for (i=0; i<NumTRBoundaryPoints; i++)
	    {
	      sum = 0.0;
	      for (j=0; j<NumLMBoundaryPoints; j++)
		if ((tmp_sum = Gaussian((float)LMBoundaryPoints[j][0],
					(float)LMBoundaryPoints[j][1],
					(float)TRBoundaryPoints[fm][i][0],
					(float)TRBoundaryPoints[fm][i][1],
					60.0))          > sum)
		  sum = tmp_sum;

	      if (sum > max_tr_sum)
		{
		  max_tr_ix = i;
		  max_tr_sum = sum;
		}
/*
	  else if (sum == max_tr_sum)
	    printf ("Equal at %f.\n", max_tr_sum);
*/
	    }

	  /* Find the LM point with the min distance (max Gaussian) to that 
	   *       TR point. 
	   */

	  max_lm_ix = 0;
	  max_lm_sum = 0.0;
	  for (i=0; i<NumLMBoundaryPoints; i++)
	    {
	      sum = Gaussian((float)LMBoundaryPoints[i][0],
			     (float)LMBoundaryPoints[i][1],
			     (float)TRBoundaryPoints[fm][max_tr_ix][0],
			     (float)TRBoundaryPoints[fm][max_tr_ix][1], 60.0);

	      if (sum > max_lm_sum)
		{
		  max_lm_ix = i;
		  max_lm_sum = sum;
		}
	    }


      
      /*  Display for debugging ...........

      XStoreName(disptr,win,"Get Theta");
      XMapWindow(disptr,win);
      gc = DefaultGC(disptr, 0);

      XSelectInput(disptr, win, 
		   ButtonPressMask | KeyPressMask | ExposureMask );
      XNextEvent(disptr, &event);
      if (event.type != Expose)
	printf (">> Not expose event.\n");

      ReadBitmapFile ("Bitmaps/LMmark",&LMmarkimage);
      ReadBitmapFile ("Bitmaps/TRmark",&TRmarkimage);


      XSetState (disptr, gc, 
		 BlackPixel(disptr,0), WhitePixel(disptr,0), 
		 GXxor, AllPlanes); 
      XPutImage (disptr, win, gc, &LMimage,
		 0, 0, LMx, LMy, 50,50);
      XPutImage (disptr, win, gc, &TRimage,
		 0, 0, TRx, TRy, 50,50);
      XPutImage (disptr, win, gc, &LMmarkimage,
		 0, 0, 
		 LMBoundaryPoints[max_lm_ix][0]-(LMmarkimage.width/2), 
		 LMBoundaryPoints[max_lm_ix][1]-(LMmarkimage.height/2), 
		 9,9);
      XPutImage (disptr, win, gc, &TRmarkimage,
		 0, 0, 
		 TRBoundaryPoints[max_tr_ix][0]-(TRmarkimage.width/2), 
		 TRBoundaryPoints[max_tr_ix][1]-(TRmarkimage.height/2), 
		 9,9);
      XFlush(disptr);


      inc = 'c';
      while (inc != ' ')
	{
	  XNextEvent(disptr, &event);
	  if (event.type == KeyPress)
	    XLookupString(&event, &inc, 1, NULL, NULL);
	  else if (event.type == Expose)
	    {  
	      XSetState (disptr, gc, 
			 BlackPixel(disptr,0), WhitePixel(disptr,0), 
			 GXxor, AllPlanes); 
	      XPutImage (disptr, win, gc, &LMimage,
			 0, 0, LMx, LMy, 50,50);
	      XPutImage (disptr, win, gc, &TRimage,
			 0, 0, TRx, TRy, 50,50);
	      XPutImage (disptr, win, gc, &LMmarkimage,
			 0, 0, 
			 LMBoundaryPoints[max_lm_ix][0]-(LMmarkimage.width/2), 
			 LMBoundaryPoints[max_lm_ix][1]-(LMmarkimage.height/2),
			 9,9);
	      XPutImage (disptr, win, gc, &TRmarkimage,
			 0, 0, 
			 TRBoundaryPoints[max_tr_ix][0]-(TRmarkimage.width/2), 
			 TRBoundaryPoints[max_tr_ix][1]-(TRmarkimage.height/2),
			 9,9);
	      XFlush(disptr);
	    }
	}
      
      XCloseDisplay( disptr );

      .........  display only for debugging purposes.   */



	  Direction((float) LMBoundaryPoints[max_lm_ix][0],
		    (float) LMBoundaryPoints[max_lm_ix][1],
		    (float) TRBoundaryPoints[fm][max_tr_ix][0],
		    (float) TRBoundaryPoints[fm][max_tr_ix][1],
		    0.0, 1.0, 
		    directionptr);

	  /*  World is upside-down (up => low y), so invert sine. */

	  ProxOrientation[fm].Sine = -directionptr->Sine;
	  ProxOrientation[fm].Cosine = directionptr->Cosine;

	}

      WriteScreenFile(SCfname);

    }
  else
    printf ("Unable to read screen file.\n");
}


/*
 *  Read in a screen file. 
 */
int ReadScreenFile (fname)
     char *fname;
{
  FILE *infile, *maofile, *fopen();
  char c;
  char tmpfilename[80];
  int i,j,k, count, ijunk, ijunk2;
  char junk[80];
  float fjunk;
  int status, success;

  infile = fopen(fname, "r");
  success = (infile != NULL);
  if (success)
    {

      /* Read in LM, first boundary, then interior. */

      fscanf (infile, "LM: %s %d %d\n", LMfilename, &LMx, &LMy);
      status = ReadBitmapFile (LMfilename,&LMimage);
      if (status != BitmapSuccess)
	{
	  printf (">> Error:  Could not read object file %s.\n", LMfilename);
	  exit(1);
	}

      /*  Find all the LM boundary points that are on. */

      NumLMBoundaryPoints = 0;
      for (i=0; i<DIM; i++)
	for (j=0; j<DIM; j++)
	  if (GetPixel(&LMimage, i,j) == 1)
	    {
	      LMBoundaryPoints[NumLMBoundaryPoints][0] = i+LMx;
	      LMBoundaryPoints[NumLMBoundaryPoints][1] = j+LMy;
	      NumLMBoundaryPoints++;
	    }

      sprintf(tmpfilename, "Filled%s", LMfilename);
      status = ReadBitmapFile (tmpfilename,&FilledLMimage);
      if (status != BitmapSuccess)
	{
	  printf (">> Error:  Could not read object file %s.\n", tmpfilename);
	  exit(1);
	}


      /* Read in TR, first boundary, then interior. */

      fscanf (infile, "TR: %s\n", TRfilename);
      status = ReadBitmapFile (TRfilename,&TRimage);
      if (status != BitmapSuccess)
	{
	  printf (">> Error:  Could not read object file %s.\n", TRfilename);
	  exit(1);
	}

      fscanf (infile, "%d", &NumFrames);
      for (i=0; i<NumFrames; i++)
	fscanf (infile, "%d %d", &TRx[i], &TRy[i]);


      /*  Find all the TR boundary points that are on. */

      NumTRBoundaryPoints = 0;
      for (i=0; i<DIM; i++)
	for (j=0; j<DIM; j++)
	  if (GetPixel(&TRimage, i,j) == 1)
	    {
	      for (k=0; k<NumFrames; k++)
		{
		  TRBoundaryPoints[k][NumTRBoundaryPoints][0] = i+TRx[k];
		  TRBoundaryPoints[k][NumTRBoundaryPoints][1] = j+TRy[k];
		}
	      NumTRBoundaryPoints++;
	    }

      sprintf(tmpfilename, "Filled%s", TRfilename);
      status = ReadBitmapFile (tmpfilename,&FilledTRimage);
      if (status != BitmapSuccess)
	{
	  printf (">> Error:  Could not read object file %s.\n", tmpfilename);
	  exit(1);
	}
      
      fclose(infile);
    }

  return ( success );
}


/*
 *  Write out a screen file. 
 */
int WriteScreenFile (fname)
     char *fname;
{
  FILE *outfile, *maofile, *fopen();
  char c;
  char tmpfilename[80];
  int i,j, count, ijunk, ijunk2;
  char junk[80];
  float fjunk;
  int status, success;

  outfile = fopen(fname, "a+");                /* Append */
  success = (outfile != NULL);
  if (success)
    {
      fprintf (outfile, "ProxOrient: \n");
      for (i=0; i<NumFrames; i++)
	fprintf (outfile, "  %f %f\n",
		 ProxOrientation[i].Sine,
		 ProxOrientation[i].Cosine);
    }

  fclose(outfile);

  return ( success );
}


/*
 *  A Gaussian.    Sigma = 60.0  seems to give about the right range for
 *    a 150x150 scene.
 */
float Gaussian(x,y, xc,yc, sigma)
     float x,y, xc,yc, sigma;
{
  /* float exp();*/
  return exp (- (((x-xc)*(x-xc)) + ((y-yc)*(y-yc))) / (sigma*sigma) );
}


/*
 *  Retrieve a pixel from an image.  Admittedly kind of ugly...  I use a
 *    function hanging off an arbitrary image (AnImage) to access pixels 
 *    in other images.  X programming at its finest.  You don't get the 
 *    function if you don't create using XCreateImage, is the problem.
 */
GetPixel(imageptr, i,j)
     XImage *imageptr;
     int i,j;
{
  return (*(AnImage->f.get_pixel)) (imageptr, i,j);
}


/*
 *  Read in a bitmap representation of an object.
 */
int ReadBitmapFile (filename,image)
     char *filename;
     XImage *image;
{
    unsigned int width, height;
    int x_hot, y_hot;
    unsigned char *data;
    int status;

    status = XmuReadBitmapDataFromFile (filename, &width, &height, &data,
					&x_hot, &y_hot);
    if (status != BitmapSuccess) return status;

    image->width = width;
    image->height = height;
    image->xoffset = 0;
    image->format = XYBitmap;
    image->data = (char *) data;
    image->byte_order = LSBFirst;
    image->bitmap_unit = 8;
    image->bitmap_bit_order = LSBFirst;
    image->bitmap_pad = 8;
    image->depth = 1;
    image->bytes_per_line = (image->width + 7) / 8;

    return BitmapSuccess;
}



/*
 *  Retrieve the value of a given pixel in the spread map.
 */
PeekSpread(i,j)
     int i,j;
{
  if ((i<0) || (i>=SDIM) || (j<0) || (j>=SDIM))
    return (0);
  else 
    return (Spread[i][j]);
}


float Distance(x,y, i,j)
     float x,y, i,j;
{
  /*float sqrt();*/

  return sqrt( ((x-i)*(x-i)) + ((y-j)*(y-j)) );
}

/*
 *  Compute direction for the input given, i.e. sin and cosine.
 *    Note that a base angle is passed in.  Sin and cos are computed
 *    relative to that angle (i.e. we really compute theta - base).
 *
 *  Recall:  sin(x-y) = sin(x) cos(y) - sin(y) cos(x)
 *           cos(x-y) = cos(x) cos(y) + sin(x) sin(y)
 */
Direction(lm_x, lm_y, tr_x, tr_y, base_sin, base_cos, directionptr)
     float lm_x, lm_y, tr_x, tr_y;
     float base_sin, base_cos;
     Angle *directionptr;
{
  int k;
  float answer;
  float xdiff, ydiff, hypot, tmp_sin, tmp_cos;

  xdiff = (tr_x - lm_x);
  ydiff = (tr_y - lm_y);
  hypot = sqrt((xdiff*xdiff) + (ydiff*ydiff));

  /*  Store theta, note minus for sin, correcting for upside-down world */

  tmp_sin = ((hypot == 0.0) ? 0.0 : ydiff / hypot);
  tmp_cos = ((hypot == 0.0) ? 0.0 : xdiff / hypot);

  /*  Subtract base from current theta */

  directionptr->Sine   = (tmp_sin*base_cos) - (base_sin*tmp_cos);
  directionptr->Cosine = (tmp_cos*base_cos) + (tmp_sin*base_sin);

  /* Roundoff */

  if ((directionptr->Sine > -0.00000001) &&
      (directionptr->Sine <  0.00000001))
    directionptr->Sine = 0.0;

  if ((directionptr->Cosine > -0.00000001) &&
      (directionptr->Cosine <  0.00000001))
    directionptr->Cosine = 0.0;
}

